package sim.app.woims;

import sim.engine.*;
import sim.util.*;
import sim.field.continuous.*;
import ec.util.*;
import java.io.*;

public /*strictfp*/ class WoimsDemo extends SimState
    {

    // dimensions of the environment
    public static final double XMIN = 0;
    public static final double XMAX = 200;
    public static final double YMIN = 0;
    public static final double YMAX = 200;

    // the diameter of each link
    public static final double DIAMETER = 1;

    // where the obstacles are located (diameter, xpos, ypos)
    public static final double[][] obstInfo = { {20, 40, 40}, {30, 135, 135} };

    // number of woims
    public static final int NUM_WOIMS = 40;

    // the difference between simulation time and woims time. it is used to compute by how much they moved. can be eliminated, but the speed
    // would need to be increased to maintain the same simulation quality
    public static final double TIMESTEP = 2;

    // for nice displaying, extra space is allocated arround the visible area
    public final static double EXTRA_SPACE = 25;

    // the maximum number of links per woim
    public final static int MAX_LINKS = 1000;

    // the woims and obstacle environments
    public Continuous2D woimsEnvironment = null;
    public Continuous2D obstaclesEnvironment = null;

    /** Creates a WoimsDemo simulation with the given random number seed. */
    public WoimsDemo(long seed)
        {
        super(new MersenneTwisterFast(seed), new Schedule(2));
        }

    public void setObjectLocation( final Woim woim, Double2D location )
        {
        // toroidal world!
        double x = (((location.x + EXTRA_SPACE - XMIN) + (XMAX-XMIN  + 2*EXTRA_SPACE)) % (XMAX-XMIN + 2*EXTRA_SPACE)) + XMIN - EXTRA_SPACE;
        double y = (((location.y + EXTRA_SPACE - YMIN) + (YMAX-YMIN  + 2*EXTRA_SPACE)) % (YMAX-YMIN + 2*EXTRA_SPACE)) + YMIN - EXTRA_SPACE;
        
        location = new Double2D( x, y );

        woimsEnvironment.setObjectLocation( woim, location );

        // to speed up the simulation, each woims knows where it is located (gets rid of a hash get call)
        woim.x = location.x;
        woim.y = location.y;
        }

    public void start()
        {
        super.start();  // clear out the schedule

        woimsEnvironment = new Continuous2D( Woim.MAX_DISTANCE, (XMAX-XMIN), (YMAX-YMIN) );
        obstaclesEnvironment = new Continuous2D( /*Math.max(Woim.numLinks*DIAMETER,30)*/30, (XMAX-XMIN), (YMAX-YMIN) );

        // Schedule the Woims -- we could instead use a RandomSequence, which would be faster,
        // but this is a good test of the scheduler
        for(int x=0;x<NUM_WOIMS;x++)
            {
            Double2D loc = null;
            Woim woim = null;
            int caz = random.nextInt(4);
            switch( caz )
                {
                case 0: loc = new Double2D( XMIN-EXTRA_SPACE,
                                            random.nextDouble()*(YMAX-YMIN-DIAMETER)+YMIN+DIAMETER/2 );
                    break;
                case 1: loc = new Double2D( XMAX+EXTRA_SPACE,
                                            random.nextDouble()*(YMAX-YMIN-DIAMETER)+YMIN+DIAMETER/2 );
                    break;
                case 2: loc = new Double2D( random.nextDouble()*(XMAX-XMIN-DIAMETER)+XMIN+DIAMETER/2,
                                            YMIN-EXTRA_SPACE );
                    break;
                case 3: loc = new Double2D( random.nextDouble()*(XMAX-XMIN-DIAMETER)+XMIN+DIAMETER/2,
                                            YMAX+EXTRA_SPACE );
                    break;
                }
            woim = new Woim();
            woimsEnvironment.setObjectLocation( woim, loc );
            woim.x = loc.x;
            woim.y = loc.y;
            schedule.scheduleRepeating(woim);
            }

        // add the obstacles to the simulation
        for( int i = 0 ; i < obstInfo.length ; i++ )
            {
            Obstacle obst = new Obstacle( obstInfo[i][0] );
            obstaclesEnvironment.setObjectLocation( obst, new Double2D( obstInfo[i][1], obstInfo[i][2] ) );
            }
        
        }

    public static void main(String[] args)
        {
        WoimsDemo woims = null;
        
        // should we load from checkpoint?  I wrote this little chunk of code to
        // check for this to give you the general idea.
        
        for(int x=0;x<args.length-1;x++)  // "-checkpoint" can't be the last string
            if (args[x].equals("-checkpoint"))
                {
                SimState state = SimState.readFromCheckpoint(new File(args[x+1]));
                if (state == null)   // there was an error -- it got printed out to the screen, so just quit
                    System.exit(1);
                else if (!(state instanceof WoimsDemo))  // uh oh, wrong simulation stored in the file!
                    {
                    System.out.println("Checkpoint contains some other simulation: " + state);
                    System.exit(1);
                    }
                else // we're ready to lock and load!  
                    woims = (WoimsDemo)state;
                }
        
        // ...or should we start fresh?
        if (woims==null)  // no checkpoint file requested
            {
            woims = new WoimsDemo(System.currentTimeMillis());  // make a new woims.  Seed the RNG with the time
            woims.start();  // prep the bugs!
            System.out.println("Starting woims.  Running for 50000 steps.");
            }
        
        long time;
        while((time = woims.schedule.time()) < 50000)
            {
            // step the schedule.  This is where everything happens.
            if (!woims.schedule.step(woims)) break;   // it won't happen that we end prematurely,
            // but it's worth checking for!

            if (time%1000==0 && time!=0) 
                System.out.println("Time Step " + time);
        
            // checkpoint
            if (time%5000==0 && time!=0)
                {
                String s = "hb." + time + ".checkpoint";
                System.out.println("Checkpointing to file: " + s);
                woims.writeToCheckpoint(new File(s));
                }
            }
            
        woims.finish();  // we don't use this, but it's good style
        }
    }
